//*****************************************************************************
//
// FILE:		main.c 
//
// DESCRIPTION: Performs tests of the CAN interface library.
//
// DOCUMENTS:	-
//
//*****************************************************************************

//*****************************************************************************
// INCLUDE FILES
//*****************************************************************************

#pragma warning( push, 3 )
#include <windows.h>
#pragma warning( pop )

#include <stddef.h>
#include <stdio.h>
#include <tchar.h>
#include "can.h"

#pragma warning( disable : 4127 )

//*****************************************************************************
// MANIFEST CONSTANTS
//*****************************************************************************
//
// Default values
//
#define DEF_NETNAME1 _TEXT("CAN1")
#define DEF_NETNAME2 _TEXT("CAN2")
#define DEF_BAUDRATE 125000
#define DEF_LOOPDELAY 10

#define	CAN_MAX_NAME_LENGTH		50
#define	INT_MAX_LENGTH			sizeof(_TEXT("4294967295"))

//
// Lock ddlink operations.
//
#define LOCK_PRINTOUT						\
  do {										\
    (void) WaitForSingleObject(				\
	    hPrintoutMutex,						\
        INFINITE );							\
 } while (0)

//
// Unlock ddlink operations.
//
#define UNLOCK_PRINTOUT						\
  do {										\
    (void) ReleaseMutex( hPrintoutMutex );	\
 } while (0)

//*****************************************************************************
// GLOBAL DATA, EMPTY!
//*****************************************************************************

//*****************************************************************************
// CONSTANTS
//*****************************************************************************

//*****************************************************************************
// TYPES
//*****************************************************************************
struct MasterThreadParam{
	BOOL continueTest;
	LPTSTR netName;
	int baudRate;
	DWORD dwLoopDelay;
	BOOL bPrintout;
};

struct SlaveThreadParam{
	BOOL continueTest;
	LPTSTR netName;
	int baudRate;
	DWORD dwLoopDelay;
	BOOL bPrintout;
	CanMsgId *pMsgSel;
};

//*****************************************************************************
// MODULE DATA
//*****************************************************************************

static CanMsgId msgSel[][5] = {
	{             0, 0, 0 ,0, 0 },
	{             1, 1, 0 ,0, 0 },
	{ (CanMsgId) -1, 1, 0 ,0, 0 },
	{             2, 2, 3 ,0, 0 },
	{ (CanMsgId) -3, 3, 4 ,5, 0 },
	{             4, 4, 5 ,6, 7 }
};

static TCHAR netName1[CAN_MAX_NAME_LENGTH], netName2[CAN_MAX_NAME_LENGTH];

static struct MasterThreadParam masterThreadParam;
static struct SlaveThreadParam slaveThreadParam[sizeof(msgSel) / sizeof(msgSel[0])];

static HANDLE hPrintoutMutex;	

//*****************************************************************************
// LOCAL FUNCTION DECLARATIONS
//*****************************************************************************

static DWORD
SystemTick(
	void
	);

static DWORD WINAPI
MasterThread(
	struct MasterThreadParam *pData
	);

static DWORD WINAPI
SlaveThread(
	struct SlaveThreadParam *pData
	);

//*****************************************************************************
// INTERFACE FUNCTIONS
//*****************************************************************************

//*****************************************************************************
//
// FUNCTION:	main
//
// DESCRIPTION:	ANSI-C standard main function.
//
// PARAMETERS:	See ANSI-C standard.
//
// RETURNS:		See ANSI-C standard.

//
//*****************************************************************************		
int
main(
	int argc,
	char** argv
	)
{	
	TCHAR intValue[INT_MAX_LENGTH];
	DWORD dwSystemTick, dwLoopDelay;
	int baudRate, i;
	BOOL bPrintout;
	TCHAR title[sizeof(netName1) + sizeof(netName2) + 50];


	HANDLE hThread[1 + sizeof(msgSel) / sizeof(msgSel[0])];
	DWORD dwThreadId[1 + sizeof(msgSel) / sizeof(msgSel[0])];

	if ( argc && argv ); // Avoid compiler warnings.

	do
	{

		//
		// Get system tick.
		//
		dwSystemTick = SystemTick();

		//
		// Get name of first CAN net.
		//
		_tprintf(_TEXT("Enter first CAN net name[%s]: "), DEF_NETNAME1);
		_fgetts(netName1, CAN_MAX_NAME_LENGTH, stdin);
		if (netName1[0] == _TEXT('\n'))
		{
			_tcscpy(netName1,DEF_NETNAME1);
		}
		else
		{
			netName1[_tcslen(netName1) - 1] = _TEXT('\0');
		}

		//
		// Get name of second CAN net.
		//
		_tprintf(_TEXT("Enter second CAN net name[%s]: "), DEF_NETNAME2);
		_fgetts(netName2, CAN_MAX_NAME_LENGTH, stdin);
		if (netName2[0] == _TEXT('\n'))
		{
			_tcscpy(netName2,DEF_NETNAME2);
		}
		else
		{
			netName2[_tcslen(netName2) - 1] = _TEXT('\0');
		}

		//
		// Get baudrate.
		//
		baudRate = DEF_BAUDRATE;
		_tprintf(_TEXT("Enter baudrate bit/s [%d]: "), DEF_BAUDRATE);
		_fgetts(intValue, INT_MAX_LENGTH, stdin);
		if (intValue[0] != _TEXT('\n'))
		{
			baudRate = _ttoi(intValue);
		}

		//
		// Get loop delay.
		// The delay is rounded to a multiple of the system tick.
		//
		dwLoopDelay = (DEF_LOOPDELAY/dwSystemTick)*dwSystemTick;
		_tprintf(_TEXT("Enter loop delay ms [%d]: "), dwLoopDelay);
		_fgetts(intValue, INT_MAX_LENGTH, stdin);
		if (intValue[0] != _TEXT('\n'))
		{
			dwLoopDelay = (_ttol(intValue)/dwSystemTick)*dwSystemTick;
			if (dwLoopDelay != (DWORD) _ttol(intValue))
				_tprintf(
					_TEXT("\tLoop delay rounded to a multiple of the system tick %d ms\n"),
					dwLoopDelay);
		}	

		//
		// Get printout flag.
		// The delay is rounded to a multiple of the system tick.
		//
		bPrintout = FALSE;
		_tprintf(_TEXT("Enable printouts [%d]: "), bPrintout);
		_fgetts(intValue, INT_MAX_LENGTH, stdin);
		if (intValue[0] != _TEXT('\n'))
		{
			bPrintout = _ttol(intValue);
		}	

		//
		// Set console window title 
		//
		_stprintf(title, _TEXT("Can ping pong %s <-> %s"),
			netName1, netName2);
		SetConsoleTitle(title);

		//
		// Create printout mutex. 
		//
		if (( hPrintoutMutex =
				CreateMutex(
					NULL,
					FALSE,
					NULL )) == NULL )
		{
			_tprintf(_TEXT(
				"!!ERROR, %lu received when calling \"CreateMutex\" ABORTING tests\n"),
				GetLastError());
			break;
		}

		//
		// Create threads that should perform ping-pong.
		//
		_tprintf(_TEXT("Creating, ping pong, master and slave thread\n"));
		masterThreadParam.continueTest = TRUE;
		masterThreadParam.netName = netName1;
		masterThreadParam.baudRate = baudRate;
		masterThreadParam.dwLoopDelay = dwLoopDelay;
		masterThreadParam.bPrintout = bPrintout;
		if ((hThread[0] =
				CreateThread(
					NULL,
					0,
					(LPTHREAD_START_ROUTINE) MasterThread,
					&masterThreadParam,
					0,
					&dwThreadId[0])) == NULL)
		{
			_tprintf(_TEXT(
				"!!ERROR, %lu received when calling \"CreateThread\" ABORTING tests\n"),
				GetLastError());
			break;
		}

		for (i = 0; i < sizeof(msgSel) / sizeof(msgSel[0]); i++ )
		{
			slaveThreadParam[i].continueTest = TRUE;
			slaveThreadParam[i].netName = netName2;
			slaveThreadParam[i].baudRate = baudRate;
			slaveThreadParam[i].dwLoopDelay = dwLoopDelay;
			slaveThreadParam[i].bPrintout = bPrintout;
			slaveThreadParam[i].pMsgSel = msgSel[i];
			if ((hThread[i+1] =
					CreateThread(
						NULL,
						0,
						(LPTHREAD_START_ROUTINE) SlaveThread,
						&slaveThreadParam[i],
						0,
						&dwThreadId[i+1])) == NULL)
			{
				_tprintf(_TEXT(
					"!!ERROR, %lu received when calling \"CreateThread\" ABORTING tests\n"),
					GetLastError());
				break;
			}
		}

		//
		// Wait for user to abort tests
		//
		_tprintf(_TEXT("Press any key to abort test"));
		(void) _gettchar();

		masterThreadParam.continueTest = FALSE;

		for (i = 0; i < sizeof(msgSel) / sizeof(msgSel[0]); i++ )
		{
			slaveThreadParam[i].continueTest = FALSE;
		}

		//
		// Wait for threads to terminate.
		//
		if (WaitForMultipleObjects( 1 + sizeof(msgSel) / sizeof(msgSel[0]),
				hThread,
				TRUE,
				5000 ) == WAIT_FAILED)
		{
			_tprintf(_TEXT(
				"!!ERROR, %lu received when waiting for threads to exit\n"),
				GetLastError());
		}

		// Close handle to threads.
		CloseHandle(hThread[0]);
		for (i = 0; i < sizeof(msgSel) / sizeof(msgSel[0]); i++ )
		{
			CloseHandle(hThread[i+1]);
		}

	} while ( 0 );

	//
	// Wait for user to close window
	//
	_tprintf(_TEXT("Press any key to close window"));
	(void) _gettchar();

	ExitProcess(0);

	return 0;		// Avoid compiler warning

}

//*****************************************************************************
// LOCAL FUNCTIONS
//*****************************************************************************

//*****************************************************************************
//
// FUNCTION:	SystemTick
//
// DESCRIPTION:	Measures the system tick
//
// PARAMETERS:	None.
//
// RETURNS:		The system tick in ms.
//
//*****************************************************************************
static DWORD
SystemTick(
	void
	)
{
	int i;
	DWORD dwMsStart, dwMsEnd;

	for (i=0; i<100; i++)
	{
		dwMsStart = GetTickCount();
		Sleep(1);
		dwMsEnd = GetTickCount();
		if (dwMsEnd-dwMsStart != 0)
			return (dwMsEnd-dwMsStart);
	}
	return 0;
}

//*****************************************************************************
//
// FUNCTION:	MasterThread
//
// DESCRIPTION:	Sends CAN messages to and receives CAN messages from
//				SlaveThread
//
// PARAMETERS:	Thread specific parameters supplied from main thread.
//
// RETURNS:		None.
//
//*****************************************************************************
static DWORD WINAPI
MasterThread(
	struct MasterThreadParam *pData
	)
{
	CanHandle hInterface;

	CanMsg canMsg;
	DWORD dwDataLength;
	ULONG i = 0, j, msgSent = 0;

	for (;;)
	{
		hInterface = CanOpen(pData->netName);
		if ( hInterface == NULL )
		{
			_tprintf(
				_TEXT("\n!!ERROR, can not open net:%s\n"),
				pData->netName );
			_tprintf(
				_TEXT("\tError %lu received when calling \"CanOpen\"\n"),
				GetLastError() );
			break;
		}

		Sleep(1000);		// Give Slave thread time to connect to can net.

		while (pData->continueTest)
		{
			i = (i + 1)%9;
			if ( !i ) i = 1;
			dwDataLength = i % ( CAN_MAX_MSG_LENGTH + 1 );
			canMsg.id = i;

			for ( j = 0; j < dwDataLength; j++ )
				canMsg.data[j] = (UCHAR) i;

			if ( !CanSend(
					hInterface,
					&canMsg,
					dwDataLength,
					FALSE ))
			{
				_tprintf(
					_TEXT("\n!!ERROR, can not send data on net:%s\n"),
					pData->netName );
				_tprintf(
					_TEXT("\tError %lu received when calling \"CanSend\"\n"),
					GetLastError() );
//				break;
			}
			else if ( pData->bPrintout )
			{
				LOCK_PRINTOUT;
				msgSent++;
				_tprintf( _TEXT("\nMessage (%lu) sent on net:%s\n"),
					msgSent, pData->netName );
				_tprintf( _TEXT("\tId: %#x\n"), canMsg.id );
				_tprintf( _TEXT("\tLength: %lu\n"), dwDataLength );
				_tprintf( _TEXT("\tData:"));
				for (i=0; i < dwDataLength; i++)
				{
					_tprintf(_TEXT(" %#2x"), canMsg.data[i] );
				}
				_tprintf( _TEXT("\n"));
				UNLOCK_PRINTOUT;
			}

			Sleep(pData->dwLoopDelay);
		}

		if ( !CanClose( hInterface ))
		{
			_tprintf(
				_TEXT("\n!!ERROR, can not close net:%s\n"),
				pData->netName );
			_tprintf(
				_TEXT("\tError %lu received when calling \"CanClose\"\n"),
				GetLastError() );
			break;
		}
		break;
	}

	ExitThread(0);

	return 0;		// Never executed, inserted to avoid compiler error.
}

//*****************************************************************************
//
// FUNCTION:	SlaveThread
//
// DESCRIPTION:	Sends CAN messages to and receives CAN messages from
//				SlaveThread
//
// PARAMETERS:	Thread specific parameters supplied from main thread.
//
// RETURNS:		None.
//
//*****************************************************************************
static DWORD WINAPI
SlaveThread(
	struct SlaveThreadParam *pData
	)
{
	CanHandle hInterface;
	DWORD noBytesRead;
	DWORD i;
	CanMsgId *pCanMsgSel;
	CanMsg canMsg;

	for (;;)
	{
		hInterface = CanOpen(pData->netName);
		if ( hInterface == NULL )
		{
			_tprintf(
				_TEXT("\n!!ERROR, can not open net:%s\n"),
				pData->netName );
			_tprintf(
				_TEXT("\tError %lu received when calling \"CanOpen\"\n"),
				GetLastError() );
			break;
		}
		pCanMsgSel =  (pData->pMsgSel[0] != 0) ? pData->pMsgSel : NULL;
		if (pCanMsgSel == NULL)
		{
			//
			// start with timeout checking.
			//
			DWORD msStart, msEnd, lastError;
			pData->pMsgSel[0] = 1;
			pData->pMsgSel[1] = 0;
			for (i = 0; i < 5; i++)
			{
				msStart = GetTickCount();
				if ( !CanReceive(
						hInterface,
						&canMsg,
						&noBytesRead,
						pData->pMsgSel,
						i*125 ))
				{
					msEnd = GetTickCount();
					if (( lastError = GetLastError() ) == ERROR_TIMEOUT )
					{
						if (pData->bPrintout )
						{
							LOCK_PRINTOUT;
							_tprintf(
								_TEXT("\n Called \"CanReceive\"\n") );
							_tprintf(
								_TEXT("\tExpected timeout value %lu ms got %lu ms\n"),
								i*125,
								(msEnd < msStart) ?
									0xffffffff - msStart + 1 + msEnd :
									msEnd - msStart );
							UNLOCK_PRINTOUT;
						}
					}
					else
					{
						_tprintf(
							_TEXT("\n!!ERROR, can not receive data on net:%s\n"),
							pData->netName );
						_tprintf(
							_TEXT("\tError %lu received when calling \"CanReceive\"\n"),
							GetLastError() );
						break;
					}
				}
				else
				{
					_tprintf(
						_TEXT("\n!!ERROR,Called \"CanReceive\"\n"),
						pData->netName );
					_tprintf( _TEXT("\tExpected timeout got success\n") );
				}
			}
		}
		while (pData->continueTest)
		{
			if ( !CanReceive(
					hInterface,
					&canMsg,
					&noBytesRead,
					pCanMsgSel,
					INFINITE ))
			{
				_tprintf(
					_TEXT("\n!!ERROR, can not receive data on net:%s\n"),
					pData->netName );
				_tprintf(
					_TEXT("\tError %lu received when calling \"CanReceive\"\n"),
					GetLastError() );
//				break;
			}
			else if ( pData->bPrintout )
			{
				LOCK_PRINTOUT;
				_tprintf( _TEXT("\nMessage received on net:%s\n"),
					pData->netName );
				_tprintf( _TEXT("\tMessage selector is:") );
				
				if (pCanMsgSel == NULL)
				{
						_tprintf( _TEXT(" NULL") );

				}
				else
				{
					for (i=0; i <= (DWORD) abs( (int) pCanMsgSel[0] ); i++)
					{
						_tprintf( _TEXT(" %d"), pCanMsgSel[i] );
					}
				}
				_tprintf( _TEXT("\n"));

				_tprintf( _TEXT("\tId: %#x\n"), canMsg.id );
				_tprintf( _TEXT("\tLength: %lu\n"), noBytesRead );
				_tprintf( _TEXT("\tData:"));
				for (i=0; i < noBytesRead; i++)
				{
					_tprintf(_TEXT(" %#2x"), canMsg.data[i] );
				}
				_tprintf( _TEXT("\n"));
				UNLOCK_PRINTOUT;
 			}
		}

		if ( !CanClose( hInterface ))
		{
			_tprintf(
				_TEXT("\n!!ERROR, can not close net:%s\n"),
				pData->netName );
			_tprintf(
				_TEXT("\tError %lu received when calling \"CanClose\"\n"),
				GetLastError() );
			break;
		}
		break;
	}

	ExitThread(0);

	return 0;		// Never executed, inserted to avoid compiler error.

}
